/* CPNix.c */ 
#include <signal.h> 
#include <setjmp.h> 
#include <sys/ioctl.h> 
#include <termio.h> 
#include <fcntl.h> 
#include <stdio.h> 
#include <ctype.h>

#define SOH       0x01
#define STX       0x02
#define ETX       0x03
#define EOT       0x04
#define ACK       0x06
#define DLE       0x10
#define NAK       0x15
#define CAN       0x18

#define RETRIES     12
#define CRC_TRIES    2
#define PADCHAR   0x1A

#define CRC        'C'

#define TRUE        1
#define FALSE       0
#define ESC        27

#define EM       0x19
char time_out;
int tty_fd;
struct termio tty_mode, /* Conterm a configuracao do terminal */
              tty_save;
jmp_buf       env;

int TIME_OUT;
int xonxoff_enabled;

int tries;                 /* Retry Counter */
unsigned char bf [135], send_char;  /* I/O Buffer  */
FILE *fp;

static char *errs[] = {
  "Timed Out         ",
  "Invalid SOH       ",
  "Invalid block #   ",
  "Invalid chksum/crc"
};

main (argv, argc)
int argv;
char *argc [];
{
  if (argv != 3) 
    uso ();
  if (!strcmp ("-t", argc [1])) {
    set_terminal ();
    send_protocol_xmodem (argc [2]);
    restore_terminal ();
  }
  else
    if (!strcmp ("-r", argc [1])) {
      set_terminal ();
      rec_protocol_xmodem (argc [2]);
      restore_terminal ();
    }
    else
      uso ();

}

int key_time_out ()
{
  longjmp (env, TRUE);
}

char wait_response (sec)
int sec;
{
  char ch;
  if (setjmp (env)) {
    time_out = TRUE;
    return ((char) EM);
  }
  signal (SIGALRM, key_time_out);
  alarm (sec);
  read (tty_fd, &ch, 1);
  time_out = FALSE;
  alarm (0);
  return (ch);
}

set_terminal ()
{
  tty_fd = open ("/dev/tty", 2);
  if (tty_fd == -1) {
    printf ("Cpnix: Erro ao abrir /dev/tty.\n");
    exit (0);
  }
  ioctl (tty_fd, TCGETA, &tty_mode);
  ioctl (tty_fd, TCGETA, &tty_save);
  tty_mode.c_iflag = (IXANY);
  tty_mode.c_oflag &= ~(077);
  tty_mode.c_lflag = 0100000;
  ioctl (tty_fd, TCSETA, &tty_mode);
}

restore_terminal ()
{
  ioctl (tty_fd, TCSETA, &tty_save);
  close (tty_fd);
}

uso ()
{
  printf ("Cpnix: Usar: cpnix -t/-r arquivo\n");
  exit (0);
}

int rec_protocol_xmodem (filename)
char *filename;
{
  int times = 0;
  int estado;
  char response;

  send_char = ESC;
  write (tty_fd, &send_char, 1);
  send_char = '[';
  write (tty_fd, &send_char, 1);
  send_char = 'R';
  write (tty_fd, &send_char, 1);
  response = wait_response (10);
  if (time_out) {
   /* printf ("Cpnix: time out\n"); */
    return FALSE;
  }
  if (response != ACK) {
   /*  printf ("Cpnix: Erro ACK esperado.\n"); */
    return FALSE; 
  }
  if (send_filename (filename) == TRUE) { 
    creat (filename);
    chmod (filename, 0777);
    if ((fp = fopen (filename, "w")) == NULL) {
    /*  printf("\nCpnix: Nao consegui abrir o arquivo %s.\n", filename); */
      return FALSE;
    }
    else {
       /* printf ("Cpnix: Preparar para receber\n"); */
      estado = receive_xmodem ();
    } 
  }
  else
    return FALSE;
  fclose (fp);
  return (estado);
}


/* receive with xmodem protocol */
int receive_xmodem(void)
{
  int blck_no, i,
      neg_blckno,
      blk    = 0,
      soh    = 0,
      crc_in = TRUE,
      fst    = TRUE,
      pula = 0;
  unsigned chksum,
           cs,
           cs1;
  char bkp [130];

  long endoffile;

  xonxoff_enabled = FALSE;
  /* - send Cs then NAKs until the sender starts sending - */
  tries = 0;
  TIME_OUT = 6;
  while (soh != SOH && tries < RETRIES)
    {
      crc_in = (tries++ < CRC_TRIES);
      send_char = crc_in ? CRC : NAK;
      write (tty_fd, &send_char, 1);
      soh = wait_response (10);
      if (!time_out && soh != SOH)
        sleep(6);
    } 
  while (tries < RETRIES)
    {
      if (time_out)  
        error(0, NAK);
      /* -- Receive the data and build the file -- */
     /*  printf("\rCpnix: Block %d   ", blk + 1); */
      if (!fst)
        {
          TIME_OUT = 10;
          soh = wait_response (10);
          if (time_out)
            continue;
          if (soh == CAN)
            break;
          if (soh == EOT)
            {
              send_char = ACK;
              write (tty_fd, &send_char, 1);
              break;
            }
        }
      fst = FALSE;
      TIME_OUT = 1;
      blck_no  = wait_response (10) & 0xFF;    /* block number   */
      neg_blckno = wait_response (10) & 0xFF;  /* 1's complement */
      chksum = 0;
      /* ---- data block ----- */
      for (i = 0; i < 129; i++)
        {
          *(bf + i) = wait_response (10); 
          if (time_out) 
            break;
          chksum = (chksum + (*(bf + i)) & 0xFF) & 0xFF;
        }
      if (time_out)
        continue;
      pula = 0;
      /* ---- checksum or crc from sender ---- */
      cs = wait_response (10) & 0xFF;
      if (crc_in)
        {
          cs1 = wait_response (10) & 0xFF;
          cs = (cs << 8) + cs1;
        }
      if (time_out)
        continue;
      if (soh != SOH)
        {       /* check the SOH */
          error(1, NAK);
          continue;
        }
      /* --- same as previous block number? --- */
      if (blck_no == blk) {
         pula = 1;
      
        /*if (fseek(fp, -128L, 1) == -1)
           fseek (fp, 0L, 0);  */
     }
        /* --- no, next sequential block number? --- */
      else
        if (blck_no != ((blk + 1) & 0xFF))
          {
            error(2, CAN);
            break;
          }
      blk = blck_no;
      /* --- test the block # 1s complement --- */
      if ((neg_blckno & 0xFF) != (~blk & 0xFF))
        {
          error(2, NAK);
          continue;
        }
      if (crc_in)
        chksum = compute_crc(bf, 130);
      /* --- test chksum or crc vs one sent --- */
      if (cs != chksum)
        {
          error(6, NAK);
          continue;
        }
      soh = blck_no = neg_blckno = cs = 0;
      tries = 0;
      /* --- write the block to disk --- */
      endoffile = (long) bf [128];
      if (pula == 0) {
        grave (fp, bf, endoffile);
      } 
      /*fwrite (bf, 1, (int) endoffile, fp); */
      send_char = ACK;
      write (tty_fd, &send_char, 1);
    }
  xonxoff_enabled = TRUE;
}

error (erno, rtn)
int erno,
    rtn;
{
  ++tries;
  if (TIME_OUT == 1)
    /* printf("\r\nCpnix: %s (%d tries)\n", errs [erno], tries); */
  send_char = rtn;
  write (tty_fd, &send_char, 1);
}

int send_protocol_xmodem (filename)
char *filename;
{
  int estado, times, response;

  if ((fp = fopen(filename, "r")) == NULL) {
    printf("Cpnix: Nao achei o arquivo %s.\n\r", filename);
    return FALSE;
  }
  send_char = ESC;
  write (tty_fd, &send_char, 1);
  send_char = '[';
  write (tty_fd, &send_char, 1);
  send_char = 'T';
  write (tty_fd, &send_char, 1);
  response = wait_response (10);
  if (time_out) {
   /* printf ("Cpnix: time out\n"); */
    return FALSE;
  }
  if (response != ACK) {
   /*  printf ("Cpnix: Erro ACK esperado.\n"); */
    return FALSE; 
  }
  if (send_filename (filename) == TRUE)  
    estado = send_xmodem ();
  else
    return FALSE;
  fclose(fp);
  return (estado);
}

int send_filename (filename)
char *filename;
{
  unsigned filecrc;
  int posi;
  short part_crc;
  char response;
  int tentativa = 0;

  do {
    tentativa++;

    filecrc = compute_crc (filename, strlen (filename));

    for (posi = 0; posi < strlen (filename); posi++)
      write (tty_fd, &filename [posi], 1);
    send_char = '*';
    write (tty_fd, &send_char, 1);

    part_crc = filecrc >> 8;
    send_char = (char) part_crc;
    write (tty_fd, &send_char, 1);
    part_crc = filecrc & 0xff;
    send_char = (char) part_crc;
    write (tty_fd, &send_char, 1);

    response = wait_response (10);
    if (response == '!') {
      restore_terminal ();
      exit (0);
    }
  } while (response != ACK && tentativa < 3);

  if (response == ACK)
    return TRUE;
  else
    return FALSE;
}

int send_xmodem (void)
{
  int i,
  chksum,
  ln = 0,
  end_of_file = FALSE,
  ans = 0,
  crc_out = 0;
  unsigned crc;
  int pula = 0;
  char blck_no = 1,
  resp = 0;

  long lido;

  xonxoff_enabled = FALSE;
  tries = 0;

  /* wait for the go-ahead from the receiver */

  TIME_OUT = 6;
  while (tries++ < RETRIES && crc_out != NAK && crc_out != CRC)
    crc_out = wait_response (10);
  TIME_OUT = 10;

  /*  send the file to the receiver ----------- */

  while (tries < RETRIES && !end_of_file && ans != CAN /*&& !time_out*/)
    {
      int i;
      /* ---- read the next data block ----- */

      /* setmem(bf, 0x80, PADCHAR); */
      if (pula == 0) {
        for (i = 0; i < 128; i++)
          bf [i] = PADCHAR;
        lido = 128l;
        end_of_file = le (fp, bf, &lido); 
        ln = (int) lido;
        if (ln == 0)
          break; 
      } 
      else
        pula = 0;
      /* printf("\rTernix: Block %d  ", blck_no); */
      chksum = 0;
      /* if (kbhit())
        if (getch() == ESC)
          {
            write_serial(CAN);
            break;
          } */
      send_char = SOH;
      write (tty_fd, &send_char, 1);
      send_char = blck_no;
      write (tty_fd, &send_char, 1);
      send_char = ~blck_no;
      write (tty_fd, &send_char, 1);
      bf [128] = lido;
      /* bf [128] = ln; */ 
      /* ------- send the data block ------ */
      for (i = 0; i < 0x81; i++)
        {
          write (tty_fd, &bf[i], 1);
          chksum += bf[i];     /* checksum calculation */
        }
      /* -- send error-correcting value (chksum or crc) -- */
      if (crc_out == NAK) {
        send_char = chksum & 0xFF;
        write (tty_fd, &send_char, 1);
      }
      else
        {
          crc = compute_crc(bf, 130);
          send_char = (crc >> 8) & 0xFF;
          write (tty_fd, &send_char, 1);
          send_char = crc & 0xFF;
          write (tty_fd, &send_char, 1);
	}

      /* Read ACK, NAK, or CAN from receiver */

      ans = wait_response (10);
      if (time_out) {
        do {
          ans = wait_response (10);
          resp++;
        } while (time_out && resp < 3);
        resp = 0;
      }
      if (ans == ACK)
        {
          blck_no++;
          tries = 0;
        }
      if (ans == NAK)
        {
          end_of_file = FALSE;


          ++tries;
          /*printf("\r\nTernix: %2d tries\n", ++tries); */

	  /* Position to previous block */

	  /*if (fseek(fp, -128L, 1) == -1)
            fseek(fp, 0L, 0); */
          
          pula = 1;
        }
    } /* while (tries < ... */
  if (end_of_file)
    {
      send_char = EOT;                 /* send the EOT    */
      write (tty_fd, &send_char, 1);
      wait_response (10);                     /* wait for an ACK */
     /* printf("\r\nTernix: Transfer Completed\n"); */
    }
  else
    /*printf("\r\nTernix: Transfer Aborted\n"); */
  xonxoff_enabled = TRUE;
}

compute_crc (bf, len)
char *bf;
int  len;
{
int i;

  long crc = 0;
  while (len--)
    {
      crc |= (*bf++) & 0xFF;
      for (i = 0; i < 8; i++)
        {
          crc <<= 1;
          if (crc & 0x1000000L)
            crc ^= 0x102100L;
        }
    }
  return (unsigned) (crc >> 8);
}
